#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""PyArrow array wrapper related class."""

from typing import Any, ClassVar, Dict, Iterator, Type, TypeVar, Union, overload

import pyarrow as pa

_A = TypeVar("_A", bound="ArrayWrapper")
_S = TypeVar("_S", bound="StructScalarWrapper")
_LS = TypeVar("_LS", bound="ListScalarWrapper")
_LA = TypeVar("_LA", bound="ListArrayWrapper")


class WrapperRegister:
    """The class decorator to connect pyarrow type and the pyarrow array wrapper.

    Arguments:
        pyarrow_type_id: The PyArrow type id.

    """

    _WRAPPERS: Dict[int, Type["ArrayWrapper"]] = {}
    _default_wrapper: Any = lambda x: x
    _default_wrapper.scalar = _default_wrapper

    def __init__(self, pyarrow_type_id: int) -> None:
        self._pyarrow_type_id = pyarrow_type_id

    def __call__(self, wrapper: Type[_A]) -> Type[_A]:
        """Connect pyarrow array wrapper with the pyarrow type id.

        Arguments:
            wrapper: The pyarrow array wrapper needs to be connected.

        Returns:
            The input wrapper class unchanged.

        """
        self._WRAPPERS[self._pyarrow_type_id] = wrapper

        return wrapper

    @classmethod
    def get(cls, pyarrow_type_id: int) -> Type["ArrayWrapper"]:
        """Get the corresponding registered pyarrow array wrapper.

        Arguments:
            pyarrow_type_id: The PyArrow type id.

        Returns:
            The corresponding registered pyarrow array wrapper.

        """
        return cls._WRAPPERS.get(pyarrow_type_id, cls._default_wrapper)


class ScalarWrapper:
    """The wrapper of pyarrow scalar.

    Arguments:
        scalar: The PyArrow scalar needs to be wrapped.

    """

    __slots__ = ("_scalar",)

    def __init__(self, scalar: pa.scalar) -> None:
        self._scalar = scalar

    @property
    def is_valid(self) -> bool:
        """The wrapper of pyarrow Scalar.is_valid method.

        Returns:
            Bool value indicating whether this scalar is None.

        """
        return self._scalar.is_valid  # type: ignore[no-any-return]

    def as_py(self) -> Any:
        """The wrapper of pyarrow Scalar.as_py method.

        Returns:
            Return this value as a Python builtin object.

        """
        return self._scalar.as_py()


class ArrayWrapper:
    """The wrapper of pyarrow array.

    Arguments:
        array: The PyArrow array needs to be wrapped.

    """

    __slots__ = ("_array",)

    scalar: ClassVar[Type[ScalarWrapper]]

    def __init__(self, array: pa.Array) -> None:
        self._array = array

    def __len__(self) -> int:
        return len(self._array)


class StructScalarWrapper(ScalarWrapper):
    """The wrapper of pyarrow StructScalar to make it case insensitive.

    Arguments:
        scalar: The PyArrow StructScalar needs to be wrapped.

    """

    __slots__ = ("_scalar", "_wrappers")

    def __init__(self, scalar: pa.StructScalar) -> None:  # pylint: disable=super-init-not-called
        self._scalar = scalar
        self._wrappers = {field.name: WrapperRegister.get(field.type.id) for field in scalar.type}

    @classmethod
    def from_wrapper(
        cls: Type[_S], scalar: pa.ListScalar, wrappers: Dict[str, Type[ArrayWrapper]]
    ) -> _S:
        """Create StructScalarWrapper instance by inputing scalar and wrappers.

        Arguments:
            scalar: The PyArrow StructScalar needs to be wrapped.
            wrappers: The wrappers of the input scalar.

        Returns:
            The StructScalarWrapper instance created by the input scalar and wrapper.

        """
        obj: _S = object.__new__(cls)
        obj._scalar = scalar
        obj._wrappers = wrappers

        return obj

    def __getitem__(self, key: str) -> ScalarWrapper:
        lower_key = key.lower()
        result = self._scalar[lower_key]
        wrapper = self._wrappers[key]

        return wrapper.scalar(result)


@WrapperRegister(pa.lib.Type_STRUCT)  # pylint: disable=c-extension-no-member
class StructArrayWrapper(ArrayWrapper):
    """The wrapper of pyarrow StructArray to make it case insensitive.

    Arguments:
        array: The PyArrow StructArray instance needs to be wrapped.

    """

    __slots__ = ("_array", "_wrappers")

    scalar = StructScalarWrapper

    def __init__(self, array: pa.StringArray) -> None:  # pylint: disable=super-init-not-called
        self._array = array
        self._wrappers = {field.name: WrapperRegister.get(field.type.id) for field in array.type}

    def __getitem__(self, index: int) -> "StructScalarWrapper":
        return StructScalarWrapper.from_wrapper(self._array[index], self._wrappers)

    def __iter__(self) -> Iterator["StructScalarWrapper"]:
        return (StructScalarWrapper.from_wrapper(item, self._wrappers) for item in self._array)

    def field(self, key: str) -> pa.Array:
        """The wrapper of pyarrow StructArray.field method.

        Arguments:
            key: The name of the field.

        Returns:
            The child array belonging to the field.

        """
        lower_key = key.lower()
        result = self._array.field(lower_key)
        return self._wrappers[lower_key](result)


class ListScalarWrapper(ScalarWrapper):
    """The wrapper of pyarrow ListScalar to make it case insensitive.

    Arguments:
        scalar: The PyArrow ListScalar instance needs to be wrapped.

    """

    __slots__ = ("_scalar", "_wrapper")

    def __init__(self, scalar: pa.ListScalar) -> None:  # pylint: disable=super-init-not-called
        self._scalar = scalar
        self._wrapper = WrapperRegister.get(scalar.type.value_type.id)

    @classmethod
    def from_wrapper(cls: Type[_LS], scalar: pa.ListScalar, wrapper: Type[ArrayWrapper]) -> _LS:
        """Create ListScalarWrapper instance by inputing scalar and wrapper.

        Arguments:
            scalar: The PyArrow ListScalar instance needs to be wrapped.
            wrapper: The wrapper of the input scalar.

        Returns:
            The ListScalarWrapper instance created by the input scalar and wrapper.

        """
        obj: _LS = object.__new__(cls)
        obj._scalar = scalar
        obj._wrapper = wrapper

        return obj

    @property
    def values(self) -> ArrayWrapper:
        """The wrapper of pyarrow ListScalar.values attr.

        Returns:
            The internal values of the pyarrow scalar.

        """
        return self._wrapper(self._scalar.values)


@WrapperRegister(pa.lib.Type_LIST)  # pylint: disable=c-extension-no-member
class ListArrayWrapper(ArrayWrapper):
    """The wrapper of pyarrow ListArray to make it case insensitive.

    Arguments:
        array: The PyArrow ListScalar instance needs to be wrapped.

    """

    __slots__ = ("_array", "_wrapper")

    scalar = ListScalarWrapper

    def __init__(self, array: pa.ListArray) -> None:  # pylint: disable=super-init-not-called
        self._array = array
        self._wrapper = WrapperRegister.get(array.type.value_type.id)

    @overload
    def __getitem__(self, index: int) -> ListScalarWrapper:
        ...

    @overload
    def __getitem__(self: _LA, index: slice) -> _LA:
        ...

    def __getitem__(self: _LA, index: Union[int, slice]) -> Union[ListScalarWrapper, _LA]:
        if isinstance(index, int):
            return ListScalarWrapper.from_wrapper(self._array[index], self._wrapper)

        return self.from_wrapper(self._array[index], self._wrapper)

    def __iter__(self) -> Iterator[ListScalarWrapper]:
        return (ListScalarWrapper.from_wrapper(item, self._wrapper) for item in self._array)

    @classmethod
    def from_wrapper(cls: Type[_LA], array: pa.ListArray, wrapper: Type[ArrayWrapper]) -> _LA:
        """Create ListScalarWrapper instance by inputing scalar and wrapper.

        Arguments:
            array: The PyArrow ListArray instance needs to be wrapped.
            wrapper: The wrapper of the input array.

        Returns:
            The ListScalarWrapper instance created by the input scalar and wrapper.

        """
        obj: _LA = object.__new__(cls)
        obj._array = array
        obj._wrapper = wrapper

        return obj



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Draft and DraftManager."""

from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple

import pyarrow as pa

from graviti.dataframe import RECORD_KEY
from graviti.exception import StatusError
from graviti.manager.branch import Branch
from graviti.manager.commit import Commit
from graviti.manager.common import ALL_BRANCHES, CURRENT_BRANCH, LIMIT, check_head_status
from graviti.manager.lazy import LazyPagingList
from graviti.manager.sheets import Sheets
from graviti.openapi import (
    commit_draft,
    create_draft,
    get_draft,
    get_draft_sheet,
    list_draft_data,
    list_draft_sheets,
    list_drafts,
    update_draft,
)
from graviti.operation import SheetOperation
from graviti.paging.factory import LazyLowerCaseFactory
from graviti.utility import check_type, convert_iso_to_datetime

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Draft(Sheets):  # pylint: disable=too-many-instance-attributes
    """The basic structure of the Graviti draft.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        number: The number of the draft.
        title: The title of the draft.
        branch: The based branch of the draft.
        state: The draft state which includes "OPEN", "CLOSED", "COMMITTED".
        parent_commit_id: The parent commit id.
        creator: The creator of the draft.
        created_at: The time when the draft is created.
        updated_at: The time of last update.
        description: The draft description.

    Attributes:
        parent: The parent of the draft.

    """

    _repr_attrs: Tuple[str, ...] = ("state", "branch", "creator", "created_at", "updated_at")

    def __init__(
        self,
        dataset: "Dataset",
        number: int,
        *,
        title: str,
        branch: str,
        state: str,
        parent_commit_id: Optional[str],
        creator: str,
        created_at: str,
        updated_at: str,
        description: str = "",
    ) -> None:
        self._dataset = dataset
        self.number = number
        self.title = title
        self.branch = branch
        self.state = state
        self.parent = None if parent_commit_id is None else Commit(dataset, parent_commit_id)
        self.creator = creator
        self.created_at = convert_iso_to_datetime(created_at)
        self.updated_at = convert_iso_to_datetime(updated_at)
        self.description = description
        self.operations: List[SheetOperation] = []

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("#{self.number}: {self.title}")'

    def _list_data(self, offset: int, limit: int, sheet_name: str) -> Dict[str, Any]:
        return list_draft_data(  # type: ignore[no-any-return]
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            draft_number=self.number,
            sheet=sheet_name,
            offset=offset,
            limit=limit,
        )["data"]

    def _list_sheets(self) -> Dict[str, Any]:
        return list_draft_sheets(
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            draft_number=self.number,
        )

    def _get_sheet(self, sheet_name: str) -> Dict[str, Any]:
        return get_draft_sheet(
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            draft_number=self.number,
            sheet=sheet_name,
            with_record_count=True,
        )

    def edit(self, title: Optional[str] = None, description: Optional[str] = None) -> None:
        """Update title and description of the draft.

        Arguments:
            title: The title of the draft.
            description: The description of the draft.

        """
        response = update_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            draft_number=self.number,
            title=title,
            description=description,
        )
        self.title = response["title"]
        self.description = response["description"]
        self.updated_at = convert_iso_to_datetime(response["updated_at"])

    def close(self) -> None:
        """Close the draft."""
        response = update_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            draft_number=self.number,
            state="CLOSED",
        )
        self.state = response["state"]
        self.updated_at = convert_iso_to_datetime(response["updated_at"])

    def commit(
        self, title: str, description: Optional[str] = None, update_dataset_head: bool = True
    ) -> Branch:
        """Commit the current draft.

        Arguments:
            title: The commit title.
            description: The commit description.
            update_dataset_head: Whether to update the dataset HEAD.

                * | True (the default value): The dataset will be updated to the committed
                    version. At this time, previous modifications to the dataset will be lost.
                * | False: The HEAD of the dataset will not be updated. This can be set if
                    the user needs to continue with some operations on the dataset.

        Returns:
            The :class:`~graviti.manager.branch.Branch` instance.

        Examples:
            The default scenario: ``update_dataset_head`` is True.

            >>> dataset = ws.datasets.get("Graviti-dataset-demo")
            >>> dataset.HEAD.name  # The version of the dataset is Branch("main").
            "main"
            >>> dataset.HEAD.commit_id
            "524d110ecae7474eaec9471f4a6c28b0"
            >>> draft = dataset.drafts.create("draft-4", branch="dev")
            >>> draft.commit("commit-4")
            Branch("dev")(
              (commit_id): '3db73ac2876a42c0bf43a0489ce1756a',
              (parent): Commit("1b21a40f03ab4cec814ec47ee0d10b24"),
              (title): 'commit-4',
              (committer): 'graviti-example',
              (committed_at): 2022-07-19 04:23:45+00:00
            )
            >>> dataset.HEAD.name  # The version of the dataset has been updated to Branch("dev").
            "dev"
            >>> dataset.HEAD.commit_id
            "3db73ac2876a42c0bf43a0489ce1756a"

            Set ``update_dataset_head`` to False.

            >>> dataset = ws.datasets.get("Graviti-dataset-demo")
            >>> dataset.HEAD.name  # The version of the dataset is Branch("main").
            "main"
            >>> dataset.HEAD.commit_id
            "524d110ecae7474eaec9471f4a6c28b0"
            >>> draft = dataset.drafts.create("draft-5", branch="dev")
            >>> draft.commit("commit-5", update_dataset_head=False)
            Branch("dev")(
              (commit_id): '781007a41d1641859c87cb00f8e32bf3',
              (parent): Commit("3db73ac2876a42c0bf43a0489ce1756a"),
              (title): 'commit-5',
              (committer): 'graviti-example',
              (committed_at): 2022-07-19 04:25:45+00:00
            )
            >>> dataset.HEAD.name  # The version of the dataset has not been updated.
            "main"
            >>> dataset.HEAD.commit_id
            "524d110ecae7474eaec9471f4a6c28b0"

        """
        branch_info = commit_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            draft_number=self.number,
            title=title,
            description=description,
        )
        branch_info["name"] = self.branch
        branch = Branch.from_response(self._dataset, branch_info)

        draft_info = get_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            draft_number=self.number,
        )
        self.state = draft_info["state"]
        self.updated_at = convert_iso_to_datetime(draft_info["updated_at"])

        if update_dataset_head:
            self._dataset._data = branch  # pylint: disable=protected-access

        return branch

    def upload(self, jobs: int = 8, quiet: bool = False) -> None:
        """Upload the local dataset to Graviti.

        Arguments:
            jobs: The number of the max workers in multi-thread upload, the default is 8.
            quiet: Set to True to stop showing the upload process bar.

        """
        modified_sheets = {name: df for name, df in self.items() if df.operations}

        self._upload_to_draft(self.number, jobs, quiet)

        for name, df in modified_sheets.items():
            factory = LazyLowerCaseFactory(
                len(df),
                LIMIT,
                partial(self._list_data, sheet_name=name),
                pa.struct([pa.field(RECORD_KEY, pa.string()), *df.schema.to_pyarrow()]),
            )
            df._refresh_data_from_factory(  # pylint: disable=protected-access)
                factory, self._dataset.object_permission_manager
            )


class DraftManager:
    """This class defines the operations on the draft in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(
        self,
        state: str,
        branch: Optional[str],
        offset: int,
        limit: int,
    ) -> Generator[Draft, None, int]:
        response = list_drafts(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            state=state,
            branch=branch,
            offset=offset,
            limit=limit,
        )

        for item in response["drafts"]:
            yield Draft(self._dataset, **item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(
        self, title: str, description: Optional[str] = None, branch: str = CURRENT_BRANCH
    ) -> Draft:
        """Create a draft.

        Arguments:
            title: The draft title.
            description: The draft description.
            branch: The branch name. The default value is the current branch of the dataset.

        Returns:
            The :class:`.Draft` instance with the given title and description.

        Raises:
            StatusError: When creating the draft without basing on a branch.

        """
        head = self._dataset.HEAD
        if branch is CURRENT_BRANCH:
            if not isinstance(head, Branch):
                raise StatusError(
                    "The current dataset is not on a branch, please checkout a branch first "
                    "or input the argument 'branch'"
                )
            branch = head.name

        response = create_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            title=title,
            branch=branch,
            description=description,
        )

        check_head_status(head, branch, response["parent_commit_id"])
        return Draft(self._dataset, **response)

    def get(self, draft_number: int) -> Draft:
        """Get the certain draft with the given draft number.

        Arguments:
            draft_number: The required draft number.

        Returns:
            The :class:`.Draft` instance with the given number.

        """
        check_type("draft_number", draft_number, int)
        response = get_draft(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            draft_number=draft_number,
        )
        return Draft(self._dataset, **response)

    def list(self, state: str = "OPEN", branch: str = ALL_BRANCHES) -> LazyPagingList[Draft]:
        """List all the drafts.

        Arguments:
            state: The draft state which includes "OPEN", "CLOSED", "COMMITTED", "ALL". The default
                value is "OPEN".
            branch: The branch name. The default value is all branches.

        Returns:
            The LazyPagingList of :class:`drafts<.Draft>` instances.

        """
        if branch is ALL_BRANCHES:
            _branch = None
        else:
            check_type("branch", branch, str)
            _branch = branch

        return LazyPagingList(
            lambda offset, limit: self._generate(state, _branch, offset, limit),
            LIMIT,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Paging module."""

from graviti.paging.factory import LazyFactory, LazyFactoryBase, LazyLowerCaseFactory
from graviti.paging.lists import MappedPagingList, PagingList, PagingListBase, PyArrowPagingList

__all__ = [
    "LazyFactory",
    "LazyFactoryBase",
    "LazyLowerCaseFactory",
    "MappedPagingList",
    "PagingList",
    "PagingListBase",
    "PyArrowPagingList",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Basic concepts of user-defined objects."""

from sys import maxsize
from typing import (
    AbstractSet,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    ValuesView,
    overload,
)

from graviti.utility.repr import ReprMixin, ReprType

_T = TypeVar("_T")
_K = TypeVar("_K")
_V = TypeVar("_V")


class UserSequence(Sequence[_T], ReprMixin):
    """UserSequence is a user-defined wrapper around sequence objects."""

    _data: Sequence[_T]

    _repr_type = ReprType.SEQUENCE

    def __len__(self) -> int:
        return self._data.__len__()

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[_T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Sequence[_T], _T]:
        return self._data.__getitem__(index)

    def __contains__(self, value: Any) -> bool:
        return self._data.__contains__(value)

    def __iter__(self) -> Iterator[_T]:
        return self._data.__iter__()

    def __reversed__(self) -> Iterator[_T]:
        return self._data.__reversed__()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._data.__eq__(other._data)

    def index(self, value: _T, start: int = 0, stop: int = maxsize) -> int:
        """Return the first index of the value.

        Arguments:
            value: The value to be found.
            start: The start index of the subsequence.
            stop: The end index of the subsequence.

        Returns:
            The First index of value.

        """
        return self._data.index(value, start, stop)

    def count(self, value: _T) -> int:
        """Return the number of occurrences of value.

        Arguments:
            value: The value to be counted the number of occurrences.

        Returns:
            The number of occurrences of value.

        """
        return self._data.count(value)


class UserMutableSequence(MutableSequence[_T], UserSequence[_T]):
    """UserMutableSequence is a user-defined wrapper around mutable sequence objects."""

    _data: MutableSequence[_T]

    _repr_type = ReprType.SEQUENCE

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[_T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[MutableSequence[_T], _T]:
        return self._data.__getitem__(index)

    @overload
    def __setitem__(self, index: int, value: _T) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[_T]) -> None:
        ...

    def __setitem__(self, index: Union[int, slice], value: Union[_T, Iterable[_T]]) -> None:
        # https://github.com/python/mypy/issues/7858
        self._data.__setitem__(index, value)  # type: ignore[index, assignment]

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._data.__delitem__(index)

    def __iadd__(self, value: Iterable[_T]) -> MutableSequence[_T]:
        return self._data.__iadd__(value)

    def insert(self, index: int, value: _T) -> None:
        """Insert object before index.

        Arguments:
            index: Position of the mutable sequence.
            value: Element to be inserted into the mutable sequence.

        """
        self._data.insert(index, value)

    def append(self, value: _T) -> None:
        """Append object to the end of the mutable sequence.

        Arguments:
            value: Element to be appended to the mutable sequence.

        """
        self._data.append(value)

    def clear(self) -> None:
        """Remove all items from the mutable sequence."""
        self._data.clear()

    def extend(self, values: Iterable[_T]) -> None:
        """Extend mutable sequence by appending elements from the iterable.

        Arguments:
            values: Elements to be Extended into the mutable sequence.

        """
        self._data.extend(values)

    def reverse(self) -> None:
        """Reverse the items of the mutable sequence in place."""
        self._data.reverse()

    def pop(self, index: int = -1) -> _T:
        """Return the item at index (default last) and remove it from the mutable sequence.

        Arguments:
            index: Position of the mutable sequence.

        Returns:
            Element to be removed from the mutable sequence.

        """
        return self._data.pop(index)

    def remove(self, value: _T) -> None:
        """Remove the first occurrence of value.

        Arguments:
            value: Element to be removed from the mutable sequence.

        """
        self._data.remove(value)


class UserMapping(Mapping[_K, _V], ReprMixin):
    """UserMapping is a user-defined wrapper around mapping objects."""

    _data: Mapping[_K, _V]

    _repr_type = ReprType.MAPPING

    def __len__(self) -> int:
        return self._data.__len__()

    def __getitem__(self, key: _K) -> _V:
        return self._data.__getitem__(key)

    def __contains__(self, key: object) -> bool:
        return self._data.__contains__(key)

    def __iter__(self) -> Iterator[_K]:
        return self._data.__iter__()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._data.__eq__(other._data)

    @overload
    def get(self, key: _K) -> Optional[_V]:  # pylint: disable=arguments-differ
        ...

    @overload
    def get(self, key: _K, default: Union[_V, _T] = ...) -> Union[_V, _T]:
        ...

    def get(self, key: _K, default: Any = None) -> Any:
        """Return the value for the key if it is in the dict, else default.

        Arguments:
            key: The key for dict, which can be any immutable type.
            default: The value to be returned if key is not in the dict.

        Returns:
            The value for the key if it is in the dict, else default.

        """
        return self._data.get(key, default)

    def items(self) -> AbstractSet[Tuple[_K, _V]]:
        """Return a new view of the (key, value) pairs in dict.

        Returns:
            The (key, value) pairs in dict.

        """
        return self._data.items()

    def keys(self) -> AbstractSet[_K]:
        """Return a new view of the keys in dict.

        Returns:
            The keys in dict.

        """
        return self._data.keys()

    def values(self) -> ValuesView[_V]:
        """Return a new view of the values in dict.

        Returns:
            The values in dict.

        """
        return self._data.values()


class UserMutableMapping(MutableMapping[_K, _V], UserMapping[_K, _V]):
    """UserMutableMapping is a user-defined wrapper around mutable mapping objects."""

    __marker = object()

    _data: MutableMapping[_K, _V]

    def __setitem__(self, key: _K, value: _V) -> None:
        self._data.__setitem__(key, value)

    def __delitem__(self, key: _K) -> None:
        self._data.__delitem__(key)

    def clear(self) -> None:
        """Remove all items from the mutable mapping object."""
        self._data.clear()

    @overload
    def pop(self, key: _K) -> _V:  # pylint: disable=arguments-differ
        ...

    @overload
    def pop(self, key: _K, default: Union[_V, _T] = ...) -> Union[_V, _T]:
        ...

    def pop(self, key: _K, default: Any = __marker) -> Any:
        """Remove specified item and return the corresponding value.

        Arguments:
            key: The key for dict, which can be any immutable type.
            default: The value to be returned if the key is not in the dict and it is given.

        Returns:
            Value to be removed from the mutable mapping object.

        """
        if default is self.__marker:
            return self._data.pop(key)

        return self._data.pop(key, default)

    def popitem(self) -> Tuple[_K, _V]:
        """Remove and return a (key, value) pair as a tuple.

        Pairs are returned in LIFO (last-in, first-out) order.

        Returns:
            A (key, value) pair as a tuple.

        """
        return self._data.popitem()

    def setdefault(self, key: _K, default: _V = None) -> _V:  # type: ignore[assignment]
        """Set the value of the item with the specified key.

        If the key is in the dict, return the corresponding value.
        If not, insert the key with a value of default and return default.

        Arguments:
            key: The key for dict, which can be any immutable type.
            default: The value to be set if the key is not in the dict.

        Returns:
            The value for key if it is in the dict, else default.

        """
        return self._data.setdefault(key, default)

    @overload
    def update(  # pylint: disable=arguments-differ
        self, __m: Mapping[_K, _V], **kwargs: _V
    ) -> None:
        ...

    @overload
    def update(  # pylint: disable=arguments-differ
        self, __m: Iterable[Tuple[_K, _V]], **kwargs: _V
    ) -> None:
        ...

    @overload
    def update(self, **kwargs: _V) -> None:  # pylint: disable=arguments-differ,signature-differs
        ...

    def update(self, __m: Any = (), **kwargs: _V) -> None:  # pylint: disable=arguments-differ
        """Update the dict.

        Arguments:
            __m: A dict object, a generator object yielding a (key, value) pair
                or other object which has a `.keys()` method.
            **kwargs: The value to be added to the mutable mapping.

        """
        self._data.update(__m, **kwargs)


class FrozenNameOrderedDict(Mapping[str, _V], ReprMixin):
    """This class is an immutable dict of ordered elements, supports searching the element by index.

    Arguments:
        items: The items need to be stored into the FrozenNameOrderedDict.

    """

    def __init__(
        self, items: Union[Iterable[Tuple[str, _V]], Mapping[str, _V], None] = None
    ) -> None:
        self._keys: List[str] = []
        self._data: Dict[str, _V] = {}

        if not items:
            return

        iterable = items.items() if isinstance(items, Mapping) else items
        for key, value in iterable:
            self._setitem(key, value)

    def __len__(self) -> int:
        return self._data.__len__()

    def __getitem__(self, key: Union[int, str]) -> _V:
        if isinstance(key, int):
            key = self._keys.__getitem__(key)

        return self._data.__getitem__(key)

    def __contains__(self, key: Any) -> bool:
        return self._data.__contains__(key)

    def __iter__(self) -> Iterator[str]:
        return self._keys.__iter__()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._keys.__eq__(other._keys) and self._data.__eq__(other._data)

    @classmethod
    def _check_key(cls, key: str) -> None:
        if not isinstance(key, str):
            raise TypeError(f'The key in "{cls.__name__}" should be a string')

    def _setitem(self, key: str, value: _V) -> None:
        self._check_key(key)

        if not self._data.__contains__(key):
            self._keys.append(key)
        self._data.__setitem__(key, value)


class NameOrderedDict(MutableMapping[str, _V], FrozenNameOrderedDict[_V]):
    """This class is a dict of ordered elements, supports searching the element by its index.

    Arguments:
        items: The items need to be stored into the NameOrderedDict.

    """

    def __setitem__(self, key: Union[int, str], value: _V) -> None:
        if isinstance(key, int):
            key = self._keys.__getitem__(key)

        self._setitem(key, value)

    def __delitem__(self, key: Union[int, str]) -> None:
        try:
            if isinstance(key, int):
                key = self._keys.pop(key)
            else:
                self._keys.remove(key)
        except ValueError:
            raise KeyError(key) from None

        self._data.__delitem__(key)

    def popitem(self) -> Tuple[str, _V]:
        """Remove and return a (key, value) pair as a tuple.

        Pairs are returned in LIFO (last-in, first-out) order.

        Raises:
            KeyError: When the dict is empty.

        Returns:
            A (key, value) pair as a tuple.

        """
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError("popitem(): dictionary is empty") from None

        return key, self.pop(key)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Parameter type releated classes."""


from typing import Any as TypingAny
from typing import ClassVar, Dict, List
from typing import Mapping as TypingMapping
from typing import Optional, Sequence, Tuple, Type, Union

from graviti.portex.base import PortexType as ClassPortexType
from graviti.portex.enum import (
    EnumValueDict,
    EnumValueList,
    EnumValues,
    EnumValueType,
    create_enum_values,
)
from graviti.portex.field import Fields as ClassFields
from graviti.portex.package import Imports


class ParameterType:
    """The base class of parameter type."""

    @staticmethod
    def check(arg: TypingAny) -> TypingAny:
        """Check the parameter type.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            The input argument unchanged.

        """
        return arg

    @staticmethod
    def load(content: TypingAny, _: Optional[Imports] = None) -> TypingAny:
        """Create an instance of the parameter type from the python content.

        Arguments:
            content: A python presentation of the parameter type.
            _: The imports of the parameter type.

        Returns:
            An instance of the parameter type.

        """
        return content

    @staticmethod
    def dump(arg: TypingAny) -> TypingAny:
        """Dump the parameter type instance into the python presentation.

        Arguments:
            arg: The parameter type instance.

        Returns:
            The python presentation of the input instance.

        """
        return arg


PType = Type[ParameterType]


class _JsonType(ParameterType):
    _type: ClassVar[Union[Type[TypingAny], Tuple[Type[TypingAny], ...]]]

    @classmethod
    def check(cls, arg: TypingAny) -> TypingAny:
        """Check the type of the input argument.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            The input argument unchanged.

        Raises:
            TypeError: When the input argument does not match the parameter type.

        """
        if not isinstance(arg, cls._type):
            raise TypeError(f"Argument should be a {cls._type}")

        return arg


class Any(ParameterType):
    """Unconstrained parameter type."""


class Boolean(_JsonType):
    """Parameter type for JSON Boolean."""

    _type = bool


class Array(_JsonType):
    """Parameter type for JSON Array."""

    _type = Sequence


class Mapping(_JsonType):
    """Parameter type for JSON object."""

    _type = TypingMapping


class Number(_JsonType):
    """Parameter type for JSON number."""

    _type = (int, float)


class Integer(_JsonType):
    """Parameter type for JSON integer."""

    _type = int


class String(_JsonType):
    """Parameter type for JSON string."""

    _type = str


class Enum(ParameterType):
    """Parameter type for Portex enum values."""

    @staticmethod
    def check(arg: TypingAny) -> EnumValues:
        """Check and transfer the parameter type.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            A list of enum values created by the input argument.

        """
        return create_enum_values(arg)

    @staticmethod
    def load(
        content: Union[Dict[int, EnumValueType], List[EnumValueType], None],
        _: Optional[Imports] = None,
    ) -> Optional[EnumValues]:
        """Create Portex EnumValues instance from python object.

        Arguments:
            content: A python list or dict representing a EnumValues.
            _: The imports of the Portex field.

        Returns:
            A Portex EnumValues instance created from the input python list or dict.

        Raises:
            TypeError: When the input enum values is not in list or dict format.

        """
        if content is None:
            return None

        if isinstance(content, dict):
            return EnumValueDict(content)

        if isinstance(content, list):
            return EnumValueList(content)

        raise TypeError("portex enum values should be a list or a dict")

    @staticmethod
    def dump(arg: EnumValues) -> Union[Dict[int, EnumValueType], List[EnumValueType]]:
        """Dump the input Portex EnumValues instance to a python list or dict.

        Arguments:
            arg: A Portex EnumValues instance.

        Returns:
            A Python list or dict representation of the Portex enum values.

        """
        return arg.to_pyobj()


class Fields(ParameterType):
    """Parameter type for Portex record fields."""

    @staticmethod
    def check(arg: TypingAny) -> ClassFields:
        """Check and transfer the parameter type.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            A :class:`Fields` instance created by the input argument.

        """
        return ClassFields(arg)

    @staticmethod
    def load(content: Optional[List[TypingAny]], imports: Optional[Imports] = None) -> ClassFields:
        """Create Portex field list instance from python list.

        Arguments:
            content: A python list representing a Portex field list.
            imports: The imports of the Portex field.

        Returns:
            A Portex field list instance created from the input python list.

        """
        if content is not None:
            return ClassFields.from_pyobj(content, imports)

        return ClassFields()

    @staticmethod
    def dump(arg: ClassFields) -> List[TypingAny]:
        """Dump the input Portex field list instance to a python list.

        Arguments:
            arg: A Portex field list instance.

        Returns:
            A Python list representation of the Portex field list.

        """
        return arg.to_pyobj()


class PortexType(ParameterType):
    """Parameter type for Portex type."""

    @staticmethod
    def check(arg: TypingAny) -> ClassPortexType:
        """Check the parameter type.

        Arguments:
            arg: The argument which needs to be checked.

        Returns:
            The input argument unchanged.

        Raises:
            TypeError: When the input argument is not a Portex type.

        """
        if not isinstance(arg, ClassPortexType):
            raise TypeError("Argument should be a Portex type")

        return arg

    @staticmethod
    def load(
        content: Optional[Dict[str, TypingAny]], imports: Optional[Imports] = None
    ) -> Optional[ClassPortexType]:
        """Create Portex type instance from python dict.

        Arguments:
            content: A python dict representing a Portex type.
            imports: The imports of the Portex type.

        Returns:
            A Portex type instance created from the input python dict.

        """
        if content is not None:
            return ClassPortexType.from_pyobj(content, imports)
        return None

    @staticmethod
    def dump(arg: ClassPortexType) -> Dict[str, TypingAny]:
        """Dump the instance to a python dict.

        Arguments:
            arg: A Portex type instance.

        Returns:
            A python dict representation of the Portex type.

        """
        return arg.to_pyobj(False)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Common tools."""

import warnings
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from graviti.manager.commit import Commit


LIMIT = 128


class DefaultValue:
    """This class defines the default value of parameters from methods in manager.

    Arguments:
        name: The name of the default value.

    """

    def __init__(self, name: str) -> None:
        self._name = name

    def __repr__(self) -> str:
        return self._name


CURRENT_COMMIT: Any = DefaultValue("CURRENT_COMMIT")
CURRENT_BRANCH: Any = DefaultValue("CURRENT_BRANCH")
ALL_BRANCHES: Any = DefaultValue("ALL_BRANCHES")


class StatusWarning(Warning):
    """This class defines the warning that the commit of dataset branch or tag is not up-to-date.

    Arguments:
       revision_type: The type of current dataset revision.
       name: The name of current dataset revision.

    """

    def __init__(self, revision_type: str, name: str):
        super().__init__()
        self._type = revision_type
        self._name = name

    def __str__(self) -> str:
        return (
            f"The commit id of {self._type} ({self._name}) of current dataset is not up-to-date. "
            "It is recommended to update the 'HEAD' of the dataset by the method "
            f"'checkout('{self._name}')'"
        )


def check_head_status(
    head: "Commit", remote_revision: str, remote_commit_id: Optional[str]
) -> None:
    """Check if the commit for the HEAD of the current dataset is up-to-date.

    Arguments:
        head: The current revision of the local dataset.
        remote_revision: The revision of the remote dataset in server.
        remote_commit_id: The commit id of the remote dataset in server.

    """
    if getattr(head, "name", None) == remote_revision and head.commit_id != remote_commit_id:
        warnings.warn(StatusWarning(head.__class__.__name__.lower(), remote_revision), stacklevel=2)
